/*
 * Copyright (c) 2012 Wayne Gray All rights reserved
 * 
 * This file is part of devboard.
 * 
 * Infinity PFM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Infinity PFM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Infinity PFM.  If not, see <http://www.gnu.org/licenses/>.
 */


package main

import (
   "main/couch"
    "fmt"
    "io/ioutil"
    "errors"
    "html/template"
    "net/http"
    "log"
    "strings"
    "regexp"
    "encoding/json"
)

const (
        COUCH_HOST = "127.0.0.1"
        COUCH_PORT = "5984"
	LEN_VIEW_PATH = len("/view/")
	LEN_SCRIPT_PATH = len("/scripts/")
	LEN_ADD_PATH = len("/add/")
	LEN_EDIT_PATH = len("/edit/")
	LEN_GET_PATH = len("/get/")
	LEN_DELETE_PATH = len("/delete/")
)

type Page struct {
	Title	string
	Body	[]byte
}

type Record struct {
        BoardItemId  string
        BoardItem []string
	TicketId string
}

type DBRecord struct {
        Id  string
        Value []string
}

type DBRecords struct {
	Rows []DBRecord
}

var titleValidator = regexp.MustCompile("^[a-zA-Z0-9]+$")
var templates = make(map[string]*template.Template)
var appTitle string;

func makeHandler(fn func(http.ResponseWriter, *http.Request, string, couch.Database), lenPath int, validate bool) http.HandlerFunc {
	return func(w http.ResponseWriter, r *http.Request) {
		
		title := r.URL.Path[lenPath:]
	
		if validate {
			if !titleValidator.MatchString(title) {
				log.Printf("validator failed for : %s", title)
				http.NotFound(w, r)
				return
			}
		}

		db, _ := couch.NewDatabase(COUCH_HOST, COUCH_PORT, "devboard")
		fn(w, r, title, db)
		
	}
}



func loadPage(title string) (*Page, error) {
	filename := title
	body, err := ioutil.ReadFile(filename)
	if err != nil {
		log.Printf("error on load page: %s, filename: %s", err, filename)
		return nil, err
	}
	return &Page{Title: title, Body: body}, nil
}

func viewHandler(w http.ResponseWriter, r *http.Request, title string, db couch.Database) {
	
	appTitle = title
	p, err := getBoardItems(db, title)
	
	if err != nil {
		p = &Page{Title: title}
	}
	
	renderTemplate(w, "board.html", p)
}

func scriptHandler(w http.ResponseWriter, r *http.Request, title string, db couch.Database) {
	p, err := loadPage("scripts/" + title + ".js")
	
	if err != nil {
		log.Printf("script handler error: %s", err)
		p = &Page{Title: title}
	} 

	fmt.Fprintf(w, "%s", p.Body)	
}

func getHandler(w http.ResponseWriter, r *http.Request, title string, db couch.Database) {
	sRecord := title
	rowVals := strings.Split(sRecord, "/")
	
	db_r := new(Record)

	_, err := db.Retrieve(rowVals[0], db_r)	

	if err != nil {
                log.Printf("failed to retrieve record: %s, id %s", err, rowVals[0])
		return
        }

	retVal, err := json.Marshal(db_r)

	if err != nil {
                log.Printf("marshal error: %s, id %s", err, rowVals[0])
		return
        }

	w.Write(retVal)

}

func addHandler(w http.ResponseWriter, r *http.Request, title string, db couch.Database) {	
	sRecord := title
	rowVals := strings.Split(sRecord, "/")
	record := Record{rowVals[0], rowVals, ""}
	db.Insert(record)
	http.Redirect(w, r, "/view/" + appTitle, http.StatusFound)
}

func editHandler(w http.ResponseWriter, r *http.Request, title string, db couch.Database) {
	sRecord := title
	rowVals := strings.Split(sRecord, "/")
	
	db_r := new(Record)

	rev, err := db.Retrieve(rowVals[0], db_r)	

	if err != nil {
                log.Printf("failed to retrieve record: %s, id %s", err, rowVals[0])
		return
        }

	newRow := rowVals[1:]
        db_r.BoardItem=newRow
	
	_, err = db.EditWith(db_r, rowVals[0], rev)

	if err!=nil {
		log.Printf("edit error: %s", err)
		return
	}	

	http.Redirect(w, r, "/view/", http.StatusFound)
	
}

func deleteHandler(w http.ResponseWriter, r *http.Request, title string, db couch.Database) {	

	sRecord := title
	rowVals := strings.Split(sRecord, "/")
	
	db_r := new(Record)

	rev, err := db.Retrieve(rowVals[0], db_r)	

	if err != nil {
                log.Printf("failed to retrieve record: %s, id %s", err, rowVals[0])
		return
        }
	
	if err := db.Delete(rowVals[0], rev); err != nil {
		log.Printf("delete failed, out of sync?: %s, id %s", err, rowVals[0])
		return
	}

	http.Redirect(w, r, "/view/", http.StatusFound)

}


func getBoardItems(db couch.Database, title string) (*Page, error){

        items := DBRecords{}

        err := db.Query("_view/boardItemView/boarditems",  nil, &items)
	l := len(items.Rows)

	if err!=nil || l==0 {
		
		log.Printf("get board items error %o", err) 

		//try alternate syntax
	        err := db.Query("_design/boardItemView/_view/boarditems",  nil, &items)

		l = len(items.Rows)

		if err!=nil {
		log.Printf("get board items error 2nd try: %o", err) 
			return nil, err;
		}
	}

	var res string
	var record DBRecord

	for i := 0; i < l; i++ {
		record = items.Rows[i]

		var desc string
		
		desc = record.Value[2]
		 
		res += record.Value[1] + "^" +
			record.Value[0] + "^" +
			desc + "^" +
			record.Id + "|" 
	}

	return &Page{Title: title, Body: []byte(res)}, nil
}

func initViews(){

	db, _ := couch.NewDatabase(COUCH_HOST, COUCH_PORT, "devboard")
	design := map[string]interface{}{}
        design["_id"] = "_design/boardItemView"
        views := map[string]interface{}{}
        design["views"] = views
        v := map[string]string{}
        views["v"] = v
        v["map"] = "function(doc) { if (doc.BoardItemId) emit(null,doc.BoardItem);} }"

	_, _, err := db.Insert(design)

	if err!=nil {
		log.Fatal(err)
	}

}

func renderTemplate(w http.ResponseWriter, tmpl string, p *Page) {
	err := templates[tmpl].ExecuteTemplate(w, tmpl, p)
	if err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
	}
}

func getArgs(w http.ResponseWriter, r *http.Request, lenPath int) (title string, err error) {
	title = r.URL.Path[lenPath:]
	if !titleValidator.MatchString(title) {
		http.NotFound(w, r)
		err = errors.New("Invalid Page Title")
	}
	return
}


func init() {

	for _, tmpl := range []string{"board.html"} {
		t := template.New(tmpl)
		tNew, err := t.ParseFiles(tmpl)
		if err != nil {
			log.Fatal(err)
		}
		templates[tmpl] = tNew;
	}
}


func main() {
	
	appTitle = ""

	http.HandleFunc("/view/", makeHandler(viewHandler, LEN_VIEW_PATH, false))
	http.HandleFunc("/get/", makeHandler(getHandler, LEN_GET_PATH, false))
	http.HandleFunc("/scripts/", makeHandler(scriptHandler, LEN_SCRIPT_PATH, true))
	http.HandleFunc("/add/", makeHandler(addHandler, LEN_ADD_PATH, false))
	http.HandleFunc("/edit/", makeHandler(editHandler, LEN_EDIT_PATH, false))
	http.HandleFunc("/delete/", makeHandler(deleteHandler, LEN_DELETE_PATH, false))
	http.ListenAndServe(":9999", nil)
}
